//============================================================================
// trimodel.cpp : triangle and triangle model class implementations
//============================================================================

#include <stdlib.h>
#include "trimodel.hpp"

// MACRO TO CHECK IF TWO VERTS ARE EQUAL (used in ConvertToSharedVert)
#define EQUAL(A,B) (A.x==B.x && A.y==B.y && A.z==B.z)

//----------------------------------------------------------------------------
// READS TRIMODEL FROM A FILE
//----------------------------------------------------------------------------
TriModel::TriModel(char* FileName)
{
  FILE* fp = fopen(FileName,"r");
  if (fp==NULL) { printf("ERROR: unable to open TriModel [%s]!\n",FileName); exit(1); }
  fscanf(fp, "%d", &NumTris);   // READ IN # OF TRIANGLES
  Tris = new Tri[NumTris];      // ALLOCATE THE TRIANGLES ARRAY
  for (int i=0; i<NumTris; i++) // READ IN THE TRIS
  {
    fscanf(fp, "%f %f %f", &Tris[i].A.x, &Tris[i].A.y, &Tris[i].A.z);
    fscanf(fp, "%f %f %f", &Tris[i].B.x, &Tris[i].B.y, &Tris[i].B.z);
    fscanf(fp, "%f %f %f", &Tris[i].C.x, &Tris[i].C.y, &Tris[i].C.z);
    fscanf(fp,"%x",&Tris[i].Color);   // READ IN "HEX" COLOR

    Tris[i].N = (Tris[i].B-Tris[i].A) / (Tris[i].C-Tris[i].A);
    Tris[i].N.Normalize();
  }
  fclose(fp);

  CalcExtents();
}

//----------------------------------------------------------------------------
// CONSTRUCTOR : CREATES TRIMODEL GIVEN A SHARED VERTEX TRIMODEL
//----------------------------------------------------------------------------
TriModel::TriModel(SharedVertTriModel &SVTM)
{
  NumTris = SVTM.NumTris;        // STORE # OF TRIANGLES
  Tris = new Tri[NumTris];       // ALLOCATE THE TRIANGLES ARRAY
  for (int i=0; i<NumTris; i++)  // LOAD TRIS INTO TriModel
  {
    Tris[i].A = SVTM.Verts[ SVTM.Tris[i].iA ];
    Tris[i].B = SVTM.Verts[ SVTM.Tris[i].iB ];
    Tris[i].C = SVTM.Verts[ SVTM.Tris[i].iC ];
    Tris[i].N = SVTM.Tris[i].N;  // COPY OVER THE TRI NORMAL
    Tris[i].Color = SVTM.Tris[i].Color;
  }

  Min = SVTM.Min;
  Max = SVTM.Max;
}

//----------------------------------------------------------------------------
// CALCULATES THE MIN AND MAX EXTENTS OF THE MODEL
//----------------------------------------------------------------------------
void TriModel::CalcExtents()
{
  Min = Max = Tris[0].A;
  for (int i=0; i<NumTris; i++)
  {
    if (Tris[i].A.x < Min.x) Min.x=Tris[i].A.x;
    else if (Tris[i].A.x > Max.x) Max.x=Tris[i].A.x;
    if (Tris[i].A.y < Min.y) Min.y=Tris[i].A.y;
    else if (Tris[i].A.y > Max.y) Max.y=Tris[i].A.y;
    if (Tris[i].A.z < Min.z) Min.z=Tris[i].A.z;
    else if (Tris[i].A.z > Max.z) Max.z=Tris[i].A.z;

    if (Tris[i].B.x < Min.x) Min.x=Tris[i].B.x;
    else if (Tris[i].B.x > Max.x) Max.x=Tris[i].B.x;
    if (Tris[i].B.y < Min.y) Min.y=Tris[i].B.y;
    else if (Tris[i].B.y > Max.y) Max.y=Tris[i].B.y;
    if (Tris[i].B.z < Min.z) Min.z=Tris[i].B.z;
    else if (Tris[i].B.z > Max.z) Max.z=Tris[i].B.z;

    if (Tris[i].C.x < Min.x) Min.x=Tris[i].C.x;
    else if (Tris[i].C.x > Max.x) Max.x=Tris[i].C.x;
    if (Tris[i].C.y < Min.y) Min.y=Tris[i].C.y;
    else if (Tris[i].C.y > Max.y) Max.y=Tris[i].C.y;
    if (Tris[i].C.z < Min.z) Min.z=Tris[i].C.z;
    else if (Tris[i].C.z > Max.z) Max.z=Tris[i].C.z;
  }
}

//----------------------------------------------------------------------------
// PRINTS OUT TRIANGLE MODEL TO A FILE (IN "RAW" TRIANGLE-SOUP FORMAT)
//----------------------------------------------------------------------------
void TriModel::PrintToFile(char* FileName)
{
  FILE* fp = fopen(FileName,"w");
  fprintf(fp, "%d\n", NumTris);
  for (int i=0; i<NumTris; i++)
  {
    fprintf(fp, "%f %f %f ", Tris[i].A.x, Tris[i].A.y, Tris[i].A.z);
    fprintf(fp, "%f %f %f ", Tris[i].B.x, Tris[i].B.y, Tris[i].B.z);
    fprintf(fp, "%f %f %f ", Tris[i].C.x, Tris[i].C.y, Tris[i].C.z);
    fprintf(fp, "%x\n", Tris[i].Color);
  }
  fclose(fp);
}

//----------------------------------------------------------------------------
// PRINTS OUT TO A VRML FILE (IN "RAW" TRIANGLE-SOUP FORMAT)
//----------------------------------------------------------------------------
void TriModel::PrintToVRML(char* FileName)
{
  FILE* fp = fopen(FileName,"w");
 fprintf(fp,"#VRML V1.0 ascii\n\n");  // HEADER
 fprintf(fp,"Separator\n{\n");        // START OF OBJECT
  for (int i=0; i<NumTris; i++)        // WRITE OUT TRIANGLES
  {
    fprintf(fp, "Coordinate3{point[%f %f %f,%f %f %f,%f %f %f]}\n",
            Tris[i].A.x, Tris[i].A.y, Tris[i].A.z,
            Tris[i].B.x, Tris[i].B.y, Tris[i].B.z,
            Tris[i].C.x, Tris[i].C.y, Tris[i].C.z);
    fprintf(fp, "IndexedFaceSet{coordIndex[0,1,2,-1]}\n");
  }
  fprintf(fp,"}\n");                   // FINISH THE VRML FILE
  fclose(fp);
}


//----------------------------------------------------------------------------
// CONSTRUCTOR: READ IN SHARED VERTEX TRIMODEL FROM FILE
//----------------------------------------------------------------------------
SharedVertTriModel::SharedVertTriModel(char* FileName)
{
  int i;
  FILE* fp = fopen(FileName, "r");
  if (fp==NULL) { printf("ERROR: unable to open SharedVertTriModel [%s]!\n",FileName); exit(1); }
  fscanf(fp, "%d", &NumVerts);       // READ IN NUM OF VERTS
  Verts = new Vec3f[NumVerts];       // ALLOC SHARED VERTEX ARRAY
  for (i=0; i<NumVerts; i++)         // READ IN THE VERTICES
    fscanf(fp, "%f %f %f", &Verts[i].x, &Verts[i].y, &Verts[i].z);
  fscanf(fp, "%d", &NumTris);        // READ IN NUM OF TRIS
  Tris = new SharedVertTri[NumTris]; // ALLOC TRIANGLE ARRAY
  for (i=0; i<NumTris; i++)          // READ IN TRIS
  {
    fscanf(fp, "%d %d %d", &Tris[i].iA, &Tris[i].iB, &Tris[i].iC);
    fscanf(fp,"%x",&Tris[i].Color);       // READ IN "HEX" COLOR

    Tris[i].N = (Verts[Tris[i].iB]-Verts[Tris[i].iA]) /
                (Verts[Tris[i].iC]-Verts[Tris[i].iA]);
    Tris[i].N.Normalize();
  }
  fclose(fp);

  CalcExtents();
}

//----------------------------------------------------------------------------
// CONVERTS TriModel TO SharedVertTriModel
//----------------------------------------------------------------------------
SharedVertTriModel::SharedVertTriModel(TriModel &TM)
{
  NumVerts = TM.NumTris * 3;         // INIT WITH MAX POSSIBLE VERTS
  NumTris = TM.NumTris;
  Verts = new Vec3f[NumVerts];      // ALLOC SHARED VERTEX ARRAY
  Tris = new SharedVertTri[NumTris]; // ALLOC TRIANGLE ARRAY

  int VertIndex=0;  // STORES "CURRENT" VERTEX BEING ADDED TO SHARED VERT ARRAY
  for (int i=0; i<NumTris; i++) // PROCESS EACH VERTEX OF EACH TRIANGLE
  {
    // CHECK IF EACH PT OF TRIANGLE IS ALREADY IN LIST, IF SO USE INDEX
    // OF THAT VERTEX (SHARED VERT), IF NOT ADD VERTEX TO VERT ARRAY
    int iVert;
    for (iVert=0; (iVert<VertIndex) && !EQUAL(TM.Tris[i].A,Verts[iVert]); iVert++);
    Tris[i].iA=iVert;
    if (iVert==VertIndex) { Verts[VertIndex]=TM.Tris[i].A; VertIndex++; }
    for (iVert=0; (iVert<VertIndex) && !EQUAL(TM.Tris[i].B,Verts[iVert]); iVert++);
    Tris[i].iB=iVert;
    if (iVert==VertIndex) { Verts[VertIndex]=TM.Tris[i].B; VertIndex++; }
    for (iVert=0; (iVert<VertIndex) && !EQUAL(TM.Tris[i].C,Verts[iVert]); iVert++);
    Tris[i].iC=iVert;
    if (iVert==VertIndex) { Verts[VertIndex]=TM.Tris[i].C; VertIndex++; }

    Tris[i].N = TM.Tris[i].N;  // COPY OVER THE NORMAL
    Tris[i].Color = TM.Tris[i].Color;
  }
  ReAllocNumVerts(VertIndex);  // VertIndex = # VERTS ADDED TO SHARED VERT ARRAY

  Min = TM.Min;
  Max = TM.Max;
}

//----------------------------------------------------------------------------
// CONVERTS SharedVertNormTriModel to SharedVertTriModel
//----------------------------------------------------------------------------
SharedVertTriModel::SharedVertTriModel(SharedVertNormTriModel &SVTMN)
{
  NumVerts = SVTMN.NumVerts;
  NumTris = SVTMN.NumTris;
  Verts = new Vec3f[NumVerts];      // ALLOC SHARED VERTEX ARRAY
  Tris = new SharedVertTri[NumTris]; // ALLOC TRIANGLE ARRAY
  int i;
  for (i=0; i<NumVerts; i++)     // COPY OVER VERTICES
    Verts[i] = SVTMN.Verts[i].V;
  for (i=0; i<NumTris; i++)      // COPY OVER TRIS
    Tris[i] = SVTMN.Tris[i];

  Min = SVTMN.Min;
  Max = SVTMN.Max;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
void SharedVertTriModel::CalcExtents()
{
  Min = Max = Verts[0];
  for (int i=1; i<NumVerts; i++)
  {
    if (Verts[i].x < Min.x) Min.x=Verts[i].x;
    else if (Verts[i].x > Max.x) Max.x=Verts[i].x;
    if (Verts[i].y < Min.y) Min.y=Verts[i].y;
    else if (Verts[i].y > Max.y) Max.y=Verts[i].y;
    if (Verts[i].z < Min.z) Min.z=Verts[i].z;
    else if (Verts[i].z > Max.z) Max.z=Verts[i].z;
  }
}

// REALLOCS VERTEX ARRAY TO NEW LENGTH (KEEPING THOSE FIRST NewNumVerts VERTICES)
//----------------------------------------------------------------------------
void SharedVertTriModel::ReAllocNumVerts(int NewNumVerts)
{
  if (NewNumVerts < NumVerts) // PROCEED ONLY IF LESS
  {
    Vec3f* NewVerts = new Vec3f[NewNumVerts];  // ALLOC NEW VERT ARRAY
    for (int i=0; i<NewNumVerts; i++)  // COPY OLD VERT ARRAY INTO NEW
      NewVerts[i] = Verts[i];
    delete[] Verts;                    // DELETE OLD VERT ARRAY
    Verts = NewVerts;                  // MAKE NEW VERT ARRAY THE CURRENT
    NumVerts = NewNumVerts;            // UPDATE VERT ARRAY LENGTH
  }
}

//----------------------------------------------------------------------------
// PRINTS OUT TO A FILE (IN SHARED VERT FORMAT)
//----------------------------------------------------------------------------
void SharedVertTriModel::PrintToFile(char* FileName)
{
  int i;
  FILE* fp = fopen(FileName,"w");
  fprintf(fp, "%d\n", NumVerts);
  for (i=0; i<NumVerts; i++)
    fprintf(fp, "%f %f %f\n", Verts[i].x, Verts[i].y, Verts[i].z);
  fprintf(fp, "%d\n", NumTris);
  for (i=0; i<NumTris; i++)
  {
    fprintf(fp, "%d %d %d ", Tris[i].iA, Tris[i].iB, Tris[i].iC);
    fprintf(fp, "%x\n", Tris[i].Color);
  }
  fclose(fp);
}

//----------------------------------------------------------------------------
// PRINTS OUT TO VRML FILE (IN SHARED VERTEX FORMAT)
//----------------------------------------------------------------------------
void SharedVertTriModel::PrintToVRML(char* FileName)
{
  int i;
  FILE* fp = fopen(FileName,"w");
  fprintf(fp,"#VRML V1.0 ascii\n\n");   // HEADER
  fprintf(fp,"Separator\n{\n");         // START OF OBJECT
  fprintf(fp, " Coordinate3{point[\n"); // WRITE OUT SHARED VERT ARRAY
  int N_1 = NumVerts-1;
  for (i=0; i<N_1; i++)
    fprintf(fp, "  %f %f %f,\n", Verts[i].x, Verts[i].y, Verts[i].z);
  fprintf(fp, "  %f %f %f\n", Verts[N_1].x, Verts[N_1].y, Verts[N_1].z);
  fprintf(fp, " ]}\n");
  fprintf(fp, " IndexedFaceSet{coordIndex[\n"); // WRITE OUT TRIANGLES
  N_1 = NumTris-1;
  for (i=0; i<N_1; i++)
    fprintf(fp, "  %d,%d,%d,-1,\n", Tris[i].iA, Tris[i].iB, Tris[i].iC);
  fprintf(fp, "  %d,%d,%d,-1\n", Tris[N_1].iA, Tris[N_1].iB, Tris[N_1].iC);
  fprintf(fp, " ]}\n");
  fprintf(fp,"}\n");                   // FINISH THE VRML FILE
  fclose(fp);
}


//----------------------------------------------------------------------------
// CONSTRUCTOR : READ IN SHARED VERTEX TRIMODEL FROM FILE
//----------------------------------------------------------------------------
SharedVertNormTriModel::SharedVertNormTriModel(char* FileName)
{
  int i;
  FILE* fp = fopen(FileName, "r");
  if (fp==NULL) { printf("ERROR: unable to open SharedVertNormTriModel [%s]!\n",FileName); exit(1); }
  fscanf(fp, "%d", &NumVerts);       // READ IN NUM OF VERTS
  Verts = new NormVert[NumVerts];    // ALLOC SHARED VERTEX ARRAY
  for (i=0; i<NumVerts; i++)         // READ IN THE VERTICES
  {
    fscanf(fp, "%f %f %f", &Verts[i].V.x, &Verts[i].V.y, &Verts[i].V.z);
    fscanf(fp, "%f %f %f", &Verts[i].N.x, &Verts[i].N.y, &Verts[i].N.z);
  }
  fscanf(fp, "%d", &NumTris);        // READ IN NUM OF TRIS
  Tris = new SharedVertTri[NumTris]; // ALLOC TRIANGLE ARRAY
  for (i=0; i<NumTris; i++)          // READ IN TRIS
  {
    fscanf(fp, "%d %d %d", &Tris[i].iA, &Tris[i].iB, &Tris[i].iC);
    fscanf(fp,"%x",&Tris[i].Color);       // READ IN "HEX" COLOR
  }
  fclose(fp);

  CalcExtents();
}

//----------------------------------------------------------------------------
// CONVERTS SharedVertTriModel TO SharedVertNormTriModel
//----------------------------------------------------------------------------
SharedVertNormTriModel::SharedVertNormTriModel(SharedVertTriModel &SVTM)
{
  NumVerts = SVTM.NumVerts;
  NumTris = SVTM.NumTris;
  Verts = new NormVert[NumVerts];          // ALLOC SHARED VERTEX ARRAY
  Tris = new SharedVertTri[NumTris];       // ALLOC TRIANGLE ARRAY

  float *NumSharing = new float[NumVerts]; // COUNTER FOR EACH VERTEX
  Vec3f AB, AC, Normal;
  int i;

  for (i=0; i<NumVerts; i++)
  {
    NumSharing[i]=0;              // INIT COUNTER
    Verts[i].V = SVTM.Verts[i];   // COPY OVER VERTICES
    Verts[i].N.Set(0,0,0);        // INIT VERT NORMAL
  }

  for (i=0; i<NumTris; i++)  // LOAD TRIS INTO TriModel
  {
    int iA,iB,iC;
    iA = Tris[i].iA = SVTM.Tris[i].iA;
    iB = Tris[i].iB = SVTM.Tris[i].iB;
    iC = Tris[i].iC = SVTM.Tris[i].iC;
    Normal = Tris[i].N = SVTM.Tris[i].N;
    Tris[i].Color = SVTM.Tris[i].Color;

    if (Normal.x!=0 || Normal.y!=0 || Normal.z!=0)
    {
      Verts[iA].N += Normal;
      Verts[iB].N += Normal;
      Verts[iC].N += Normal;
      NumSharing[iA]++;
      NumSharing[iB]++;
      NumSharing[iC]++;
    }
  }

  for (i=0; i<NumVerts; i++)
  {
    if (NumSharing[i] > 1)
      Verts[i].N /= NumSharing[i];
    if (NumSharing[i] > 0)
      Verts[i].N.Normalize();
  }

  delete[] NumSharing;

  Min = SVTM.Min;
  Max = SVTM.Max;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
void SharedVertNormTriModel::CalcExtents()
{
  Min = Max = Verts[0].V;
  for (int i=1; i<NumVerts; i++)
  {
    if (Verts[i].V.x < Min.x) Min.x=Verts[i].V.x;
    else if (Verts[i].V.x > Max.x) Max.x=Verts[i].V.x;
    if (Verts[i].V.y < Min.y) Min.y=Verts[i].V.y;
    else if (Verts[i].V.y > Max.y) Max.y=Verts[i].V.y;
    if (Verts[i].V.z < Min.z) Min.z=Verts[i].V.z;
    else if (Verts[i].V.z > Max.z) Max.z=Verts[i].V.z;
  }
}

//----------------------------------------------------------------------------
// PRINTS OUT TO A FILE (IN SHARED VERT FORMAT)
//----------------------------------------------------------------------------
void SharedVertNormTriModel::PrintToFile(char* FileName)
{
  int i;
  FILE* fp = fopen(FileName,"w");
  fprintf(fp, "%d\n", NumVerts);
  for (i=0; i<NumVerts; i++)
  {
    fprintf(fp, "%f %f %f  ", Verts[i].V.x, Verts[i].V.y, Verts[i].V.z);
    fprintf(fp, "%f %f %f\n", Verts[i].N.x, Verts[i].N.y, Verts[i].N.z);
  }
  fprintf(fp, "%d\n", NumTris);
  for (i=0; i<NumTris; i++)
  {
    fprintf(fp, "%d %d %d", Tris[i].iA, Tris[i].iB, Tris[i].iC);
    fprintf(fp, "%x\n", Tris[i].Color);
  }
  fclose(fp);
}

//----------------------------------------------------------------------------
// PRINTS OUT TO VRML FILE (IN SHARED VERTEX FORMAT W/ VERT NORMALS)
//----------------------------------------------------------------------------
void SharedVertNormTriModel::PrintToVRML(char* FileName)
{
  int i;
  FILE* fp = fopen(FileName,"w");
  fprintf(fp,"#VRML V1.0 ascii\n\n");   // HEADER
  fprintf(fp,"Separator\n{\n");         // START OF OBJECT

  fprintf(fp, " Coordinate3{point[\n"); // WRITE OUT SHARED VERT ARRAY
  int N_1 = NumVerts-1;
  for (i=0; i<N_1; i++)
    fprintf(fp, "  %f %f %f,\n", Verts[i].V.x, Verts[i].V.y, Verts[i].V.z);
  fprintf(fp, "  %f %f %f\n", Verts[N_1].V.x, Verts[N_1].V.y, Verts[N_1].V.z);
  fprintf(fp, " ]}\n");

  fprintf(fp, " Normal{vector[\n"); // WRITE OUT SHARED VERT ARRAY
  for (i=0; i<N_1; i++)
    fprintf(fp, "  %f %f %f,\n", Verts[i].N.x, Verts[i].N.y, Verts[i].N.z);
  fprintf(fp, "  %f %f %f\n", Verts[N_1].N.x, Verts[N_1].N.y, Verts[N_1].N.z);
  fprintf(fp, " ]}\n");

  fprintf(fp, " IndexedFaceSet{coordIndex[\n"); // WRITE OUT TRIANGLES
  N_1 = NumTris-1;
  for (i=0; i<N_1; i++)
    fprintf(fp, "  %d,%d,%d,-1,\n", Tris[i].iA, Tris[i].iB, Tris[i].iC);
  fprintf(fp, "  %d,%d,%d,-1\n", Tris[N_1].iA, Tris[N_1].iB, Tris[N_1].iC);
  fprintf(fp, " ]}\n");
  fprintf(fp,"}\n");                   // FINISH THE VRML FILE

  fclose(fp);
}

